package hr.fer.ga.imgcompression;
/**
 * 
 */
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import org.jgap.Chromosome;
import org.jgap.Configuration;
import org.jgap.Gene;
import org.jgap.Genotype;
import org.jgap.IChromosome;
import org.jgap.IGeneticOperatorConstraint;
import org.jgap.IUniversalRateCalculator;
import org.jgap.InvalidConfigurationException;
import org.jgap.Population;
import org.jgap.RandomGenerator;
import org.jgap.impl.CrossoverOperator;

/**
 *
 */
public class CrossoverOperator1 extends CrossoverOperator {
	
	/**
	 * vjerojatnost krianja zadana brojem jedinki koje
	 * e se kriati (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private int m_crossoverRate;
	
	/**
	 * vjerojatnost krianja zadana postotkom jedinki populacije
	 * (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private double m_crossoverRatePercent;

	/**
	 * kalkulator za dinamiko odreivanje vjerojatnosti krianja
	 * (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private IUniversalRateCalculator m_crossoverRateCalc;
	
	/**
	 * ako je vrijednost postavljena na true kria i novo stvorene kromosome iz
	 * populacije potomaka (kromosome stare 0 godina 
	 */
	private boolean m_xoverNewAge;
	
	public CrossoverOperator1()
		throws InvalidConfigurationException {
		super(Genotype.getStaticConfiguration());
		init();
	}
	
	public CrossoverOperator1(final Configuration config) throws InvalidConfigurationException {
		super(config);
		init();
	}
		
	@Override
	protected void init() {
	    m_crossoverRate = 6;
	    m_crossoverRatePercent = -1;
	    setCrossoverRateCalc(null);
	    setXoverNewAge(true);
	}
	
	public CrossoverOperator1(final Configuration a_configuration,
				final IUniversalRateCalculator a_crossoverRateCalculator)
										throws InvalidConfigurationException {
		super(a_configuration);
		setCrossoverRateCalc(a_crossoverRateCalculator);
	}
	
	/**
	 * Postavlja kalkulator za dinamiko odreivanje vjerojatnosti krianja
	 * 
	 * @param a_crossoverRateCalculator
	 */
	private void setCrossoverRateCalc(final IUniversalRateCalculator
											a_crossoverRateCalculator) {
		m_crossoverRateCalc = a_crossoverRateCalculator;
		if (a_crossoverRateCalculator != null) {
			m_crossoverRate = -1;
			m_crossoverRatePercent = -1d;
		}
	}
	
	public CrossoverOperator1(final Configuration a_configuration,
			final double a_crossoverRatePercentage,
			final boolean a_xoverNewAge)
				throws InvalidConfigurationException {
		super(a_configuration);
		
		if (a_crossoverRatePercentage <= 0.0d) {
			throw new IllegalArgumentException("Crossover rate must be greater zero");
		}
		m_crossoverRatePercent = a_crossoverRatePercentage;
		m_crossoverRate = -1;
		setCrossoverRateCalc(null);
		setXoverNewAge(a_xoverNewAge);
	}
	
	/**
	 * Vraa lokus gena u kromosomu koji sadri zadani alel
	 * 
	 * @param gene
	 * @return		lokus gena ili -1 ako gen s zadanim alelom ne postoji
	 */
	private int getGeneLocus(PixelGene gene, IChromosome chrom) {
		Gene[] genes = chrom.getGenes();
		for(int i = 0; i < genes.length; i++) {
			if(gene.compareTo((PixelGene)genes[i]) == 0) {
				return i;
			}
		}
		return -1;
	}
	
	/**
	 * 
	 * @param chrom
	 * @param startLocus	ukljuivo
	 * @param endLocus		ukljuivo
	 * @return
	 */
	private List<PixelGene> getGenes(IChromosome chrom, int startLocus, int endLocus) {
		Gene[] genes = chrom.getGenes();
		List<PixelGene> geneList = new ArrayList<PixelGene>();
		
		if(startLocus < 0)
			startLocus = 0;
		
		if(endLocus >= chrom.size())
			endLocus = chrom.size() - 1;
		
		for(int i = startLocus; i <= endLocus; i++) {
			geneList.add((PixelGene)genes[i]);
		}
		return geneList;
	}
	
	/**
	 * Metoda kria dva odabrana kromosoma
	 * 
	 * @param parent1
	 * @param parent2
	 * @return
	 */
	public IChromosome cross(IChromosome parent1, IChromosome parent2)
							throws InvalidConfigurationException {
		System.out.print("Kriam kromosome (operator1) " 
				+ parent1.getFitnessValue() + " i " + parent2.getFitnessValue() + " ...");
		Configuration conf = getConfiguration();
		int xp1 = getConfiguration().getRandomGenerator().nextInt(parent1.size());
		PixelGene x = (PixelGene)parent1.getGene(xp1);
		int xp2 = getGeneLocus(x, parent2);
		
		if(xp2 == -1) {
			throw new IllegalStateException("Kromosom recesivnog roditelja ne sadri jedan od gena!");
		}
		
		int maxXw = Math.abs(parent1.size() - xp2);
		if(xp2 < maxXw) {
			maxXw = xp2;
		}
		int xw = getConfiguration().getRandomGenerator().nextInt(parent1.size());
		/*
		System.out.println("parametri: ");
		System.out.println("	xp1: " + xp1);
		System.out.println("	xp2: " + xp2);
		System.out.println("	xw: " + xw);
			*/	
		List<PixelGene> X = getGenes(parent2, xp2 - xw, xp2 + xw);
		
		PixelGene currentGene;
		PixelGene[] childGenes = new PixelGene[parent1.size()];
		int childLocus = 0;
		
		/* 1. korak */
		for(int i = 0; i < xp1; i++) {
			currentGene = (PixelGene)parent1.getGene(i);
			if(!X.contains(currentGene)) {
				childGenes[childLocus] = currentGene;
				childLocus++;
			}
		}
		
		/* 2. korak */
		for(int i = 0; i < parent1.size(); i++) {
			currentGene = (PixelGene)parent1.getGene(i);
			if(X.contains(currentGene)) {
				childGenes[childLocus] = currentGene;
				childLocus++;
			}
		}
		
		/* 3. korak */
		for(int i = xp1; i < parent1.size(); i++) {
			currentGene = (PixelGene)parent1.getGene(i);
			if(!X.contains(currentGene)) {
				childGenes[childLocus] = currentGene;
				childLocus++;
			}
		}
		System.out.println("gotov");
		
		return new Chromosome(conf, childGenes);
	}
	
	/**
	 * Metoda vraa broj krianja koji je potrebno obaviti
	 * prema trenutnim postavkama.
	 * 
	 * @param populationSize
	 * @return
	 */
	private int getNumberOfCrossovers(int size) {
	    int numCrossovers = 0;
	    if (m_crossoverRate >= 0) {
	    	numCrossovers = size / m_crossoverRate;
	    } else if (m_crossoverRateCalc != null) {
	    	numCrossovers = size / m_crossoverRateCalc.calculateCurrentRate();
	    } else {
	    	numCrossovers = (int) (size * m_crossoverRatePercent);
	    }
	    return numCrossovers;
	}
	
	/**
	 * Ova metoda zove se jednom po evoluciji: mora evoluirat sve
	 * 
	 * kod evolucije: prvo zove Operatore krianja (genetic operators)
	 * onda zove prirodne selektore
	 * 
	 * @param	population 	stara populacija
	 * @param	chromosomes		izgenerirani potomci se stavljaju tu
	 */
	@Override
	public void operate(Population population, List chromosomes) {
	    int size = Math.min(getConfiguration().getPopulationSize(),
                population.size());
		int numCrossovers = getNumberOfCrossovers(population.size());
	    RandomGenerator generator = getConfiguration().getRandomGenerator();
	    IGeneticOperatorConstraint constraint = getConfiguration().
	        getJGAPFactory().getGeneticOperatorConstraint();
	    	    
	    int index1, index2;
	    for (int i = 0; i < numCrossovers; i++) {
	    	index1 = generator.nextInt(size);
	    	index2 = generator.nextInt(size);
	    	IChromosome parent1 = population.getChromosome(index1);
	    	IChromosome parent2 = population.getChromosome(index2);

	    	if (!isXoverNewAge() && parent1.getAge() < 1 && parent2.getAge() < 1) {
	    		continue;
	    	}
	    	if (constraint != null) {
	    		List v = new Vector();
	    		v.add(parent1);
	    		v.add(parent2);
	        	if (!constraint.isValid(population, v, this)) {
	        		/* zabranjen je ovaj tip krianja */
	        		continue;
	        	}
	    	}
	    	
	    	IChromosome child = null;
	    	try {
	    		child = cross(parent1, parent2);
	    	} catch(InvalidConfigurationException e) {
	    		throw new IllegalStateException();
	    	}
	    	chromosomes.add(child);
	    }
	}
}
