package hr.fer.ga.imgcompression;
/**
 * 
 */
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jgap.BreederBase;
import org.jgap.BulkFitnessFunction;
import org.jgap.Configuration;
import org.jgap.FitnessFunction;
import org.jgap.IChromosome;
import org.jgap.IInitializer;
import org.jgap.InvalidConfigurationException;
import org.jgap.Population;
import org.jgap.event.GeneticEvent;


/**
 * 
 * @author Ivan Kut
 *
 */
public class ImageBreeder extends BreederBase {

	/** 
	 * broj "elitnih" jedinki koje se biraju kao najbolje od roditelja i potomaka
	 * i prenose u sljedeu generaciju
	 */
	private int elites;
	
	/**
	 * zadnja konfiguracija koju je <code>ImageBreeder</code> koristio
	 */
	private transient Configuration m_lastConf;

	/**
	 * zadnja populacija koju je <code>ImageBreeder</code> uzgojio
	 */
	private transient Population m_lastPop;

	/**
	 * Stvara novi primjerak klase <code>ImageBreeder</code>.
	 * 
	 * @param elites	broj najboljih jedinki koje se prenose iz 
	 * 					generacije roditelja i potomaka u sljedeu
	 * 					generaciju
	 */
	public ImageBreeder(int elites) {
		super();
		this.elites = elites;
	}
	
	@Override
	public Population evolve(Population population, Configuration a_conf) {
	    Population oldPopulation = population;
	    Population offspring = null;
	    Population newPopulation = null;
	    int originalPopSize = a_conf.getPopulationSize();
    
	    /* Ako se radi o prvoj generaciji postavi starost na jedan */
	    if (a_conf.getGenerationNr() == 0) {
	      int size = oldPopulation.size();
	      for (int i = 0; i < size; i++) {
	        IChromosome chrom = oldPopulation.getChromosome(i);
	        chrom.increaseAge();
	      }
	    }
	    
	    /* updateaj fitness svih kromosoma */
	    updateChromosomes(oldPopulation, a_conf);

	    offspring = applyNaturalSelectors(a_conf, oldPopulation, true);
/*		System.out.println("offspring: ");
		List<IChromosome> chromosomes;
		chromosomes = offspring.getChromosomes();
		
		for(IChromosome chrom : chromosomes) {
			System.out.println("Kromosom: ");
			System.out.println("age: " + chrom.getAge());
			for(int i = 0; i < chrom.size(); i++) {
				System.out.println((PixelAllele)chrom.getGene(i).getAllele());
			}
		} */
	    
	    applyGeneticOperators(a_conf, offspring);
	/*	System.out.println("offspring after operators: ");
		chromosomes = offspring.getChromosomes();
		
		for(IChromosome chrom : chromosomes) {
			System.out.println("Kromosom: ");
			System.out.println("age: " + chrom.getAge());
			for(int i = 0; i < chrom.size(); i++) {
				System.out.println((PixelAllele)chrom.getGene(i).getAllele());
			}
		} */
	    
	    // Reset fitness value of genetically operated chromosomes.
	    // Normally, this should not be necessary as the Chromosome class
	    // initializes each newly created chromosome with
	    // FitnessFunction.NO_FITNESS_VALUE. But who knows which Chromosome
	    // implementation is used...
	    // ----------------------------------------------------------------
	    int currentPopSize = oldPopulation.size();
	    for (int i = originalPopSize; i < currentPopSize; i++) {
	      IChromosome chrom = oldPopulation.getChromosome(i);
	      chrom.setFitnessValueDirectly(FitnessFunction.NO_FITNESS_VALUE);
	      // Mark chromosome as new-born.
	      // ----------------------------
	      chrom.resetAge();
	      // Mark chromosome as being operated on.
	      // -------------------------------------
	      chrom.increaseOperatedOn();
	    }
	    // Increase age of all chromosomes which are not modified by genetic
	    // operations.
	    // -----------------------------------------------------------------
	    int size = Math.min(originalPopSize, currentPopSize);
	    for (int i = 0; i < size; i++) {
	      IChromosome chrom = oldPopulation.getChromosome(i);
	      chrom.increaseAge();
	      // Mark chromosome as not being operated on.
	      // -----------------------------------------
	      chrom.resetOperatedOn();
	    }
	    // Ensure fitness value of all chromosomes is udpated.
	    // ---------------------------------------------------
	    updateChromosomes(oldPopulation, a_conf);
	    // Apply certain NaturalSelectors after GeneticOperators have been applied.
	    // ------------------------------------------------------------------------
	    oldPopulation = applyNaturalSelectors(a_conf, oldPopulation, false);
	    // If a bulk fitness function has been provided, call it.
	    // ------------------------------------------------------
	    BulkFitnessFunction bulkFunction = a_conf.getBulkFitnessFunction();
	    if (bulkFunction != null) {
	      /**@todo utilize jobs: bulk fitness function is not so important for a
	       * prototype! */
	      bulkFunction.evaluate(oldPopulation);
	    }
	    // Fill up population randomly if size dropped below specified percentage
	    // of original size.
	    // ----------------------------------------------------------------------
	    if (a_conf.getMinimumPopSizePercent() > 0) {
	      int sizeWanted = a_conf.getPopulationSize();
	      int popSize;
	      int minSize = (int) Math.round(sizeWanted *
	                                     (double) a_conf.getMinimumPopSizePercent()
	                                     / 100);
	      popSize = oldPopulation.size();
	      if (popSize < minSize) {
	        IChromosome newChrom;
	        IChromosome sampleChrom = a_conf.getSampleChromosome();
	        Class sampleChromClass = sampleChrom.getClass();
	        IInitializer chromIniter = a_conf.getJGAPFactory().
	            getInitializerFor(sampleChrom, sampleChromClass);
	        while (oldPopulation.size() < minSize) {
	          try {
	            /**@todo utilize jobs as initialization may be time-consuming as
	             * invalid combinations may have to be filtered out*/
	            newChrom = (IChromosome) chromIniter.perform(sampleChrom,
	                sampleChromClass, null);
	            oldPopulation.addChromosome(newChrom);
	          } catch (Exception ex) {
	            throw new RuntimeException(ex);
	          }
	        }
	      }
	    }
	    
	    /* dodaj "elitne" u novu populaciju */
	    Population oldAndOffspring = null;
	    try {
	    	oldAndOffspring = new Population(a_conf);
	    } catch(InvalidConfigurationException e) {
	    	throw new IllegalStateException(e.getMessage());
	    }
	    oldAndOffspring.addChromosomes(oldPopulation);
	    oldAndOffspring.addChromosomes(offspring);
	    
	    IChromosome[] fittest = new IChromosome[elites]; 
	    fittest = (IChromosome[])oldAndOffspring.determineFittestChromosomes(elites)
	    				.toArray(fittest); //sta ako nema uopce dovoljno kromosoma u populaciji?  	
	/*	System.out.println("fittest: ");

		for(IChromosome chrom : fittest) {
			System.out.println("Kromosom: ");
			System.out.println("age: " + chrom.getAge());
			for(int i = 0; i < chrom.size(); i++) {
				System.out.println((PixelAllele)chrom.getGene(i).getAllele());
			}
		} */
	    
	    try {
	    	newPopulation = new Population(a_conf);
	    } catch(InvalidConfigurationException e) {
	    	throw new IllegalStateException(e.getMessage());
	    }
	    reAddFittest(newPopulation, fittest);
	    
	    /* dodaj ostatak iz potomaka */
	    int randomIndex;
	    int maxValue = offspring.size();
	    for(int i = 0; i < a_conf.getPopulationSize() - elites; i++) {
	    	randomIndex = a_conf.getRandomGenerator().nextInt(maxValue);
	    	newPopulation.addChromosome(offspring.getChromosome(randomIndex));
	    }
	    
	    /* poveaj broj generacija */
	    a_conf.incrementGenerationNr();
	    /* opali dogaaj koji oznaava da se desila evolucija */
	    m_lastPop = oldPopulation;
	    m_lastConf = a_conf;
	    a_conf.getEventManager().fireGeneticEvent(
	        new GeneticEvent(GeneticEvent.GENOTYPE_EVOLVED_EVENT, this));
	    return oldPopulation;
	}
	
	/**
	 * Metoda dodaje "elitne" kromosome u novu populaciju.
	 * 
	 * @param population
	 * @param elites
	 */
	protected void reAddFittest(Population population, IChromosome[] elites) {
		if(elites == null) {
			throw new NullPointerException("");
		}
		
		for(int i = 0; i < elites.length; i++) {
			if(!population.contains(elites[i]))
				population.addChromosome(elites[i]);
		}
	}
	
	/**
	 * Metoda radi update fitness-a svih kromosoma
	 * 
	 * @param a_pop
	 * @param a_conf
	 */
	protected void updateChromosomes(Population a_pop, Configuration a_conf) {
		int currentPopSize = a_pop.size();
	    
		/* osiguraj da su fitnessi svih kromosoma up to date */
		BulkFitnessFunction bulkFunction = a_conf.getBulkFitnessFunction();
		boolean bulkFitFunc = (bulkFunction != null);
		if (!bulkFitFunc) {
			for (int i = 0; i < currentPopSize; i++) {
				IChromosome chrom = a_pop.getChromosome(i);
				chrom.getFitnessValue();
			}
		}
	}

	public Configuration getLastConfiguration() {
		return m_lastConf;
	}

	public Population getLastPopulation() {
		return m_lastPop;
	}
	
	@Override
	public Object clone() {
		return new ImageBreeder(elites);
	}
}
