package hr.fer.ga.imgcompression;
/**
 * 
 */
import hr.fer.ga.imgcompression.clustering.FixedClusterer;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import javax.imageio.ImageIO;

import org.jgap.Chromosome;
import org.jgap.Configuration;
import org.jgap.Gene;
import org.jgap.Genotype;
import org.jgap.IChromosome;
import org.jgap.InvalidConfigurationException;
import org.jgap.Population;
import org.jgap.impl.DefaultConfiguration;



/**
 * Klasa koja vri kompresiju slika. 
 * (uitavanje slika, ureena reprezentacija, rasporeivanje u razrede,
 *  spremanje najbolje jedinke)
 * 
 * @author Ivan Kut
 * 
 * Moe kompresirati BMP, JPG i PNG slike s 24-bitnom bojom.
 * Slike sprema u PNG formatu.
 * 
 * Ova klasa koristi JGAP (Java genetic algorithms package) library:
 * http://jgap.sourceforge.net/
 *
 */
public class ImageCompressor {

	/** lista alela koji predstavljaju pixele uitane slike */
	private List<PixelGene> geneList;
	
	/** irina slike koja je uitana za kompresiju u pixelima */
	private int width;
	
	/** visina slike koja je uitana za kompresiju u pixelima */
	private int height;
	
	/** oznaava da li je slika za kompresiju uitana */
	private boolean loaded;
	
	/** oznaava da li da se spremaju informacije o svim generacijama */
	private boolean storeGenerations;
	
	/** oznaava odredini direktorij gdje e se spremati generacije */
	private String destination;
	
	/** generator nasuminih brojeva */
	private static Random randomNumberGenerator;
	
	/** najbolja jedinka nastala evolucijom kao rjeenje */
	private IChromosome solution;
	
	/** current configuration */
	Configuration config;
	
	/** klasa koja vri klasteriranje objekata */
	private static FixedClusterer clusterer = new FixedClusterer();
	
	/** broj razreda u koje se rasporeuju objekti prilikom kompresije; odreuje stupanj kompresije */
	private int numberOfClusters;
	
	/** veliina populacije */
	private int populationSize = 10;
	
	/** broj generacija */
	private int numberOfGenerations = 5;
	
	/** broj elitnih jedinki koje se kao najbolje prenose u sljedeu generaciju */
	private int numberOfElites = 4;
	
	/** parametar konstruktora C */
	private int k = 400;
	
	/** vjerojatnost krianja operatorom 1 */
	private double operator1Probability = 0.1;
	
	/** vjerojatnost krianja operatorom 2 */
	private double operator2Probability = 0.1;
	
	/** vjerojatnost mutacije */
	private double mutationProbability = 0.01;
	
	/** odredina datoteka gdje sprema kompresiranu sliku */
	private File destinationImage = new File("result.jpg");
	
	/**
	 *  doputena udaljenost dvaju pixela a da se ne treitra kao greka
	 *  pri usporeivanju kompresirane slike i originala 
	 */
	private int errorThreshold = 5;
		
	static {
		randomNumberGenerator = new Random(System.currentTimeMillis());
	}
	
	/**
	 * Konstruktor.
	 */
	public ImageCompressor() {
		init();
	}
	
	/** 
	 * Postavlja defaultne postavke
	 */
	private void init() {
		config = new DefaultConfiguration();
		numberOfClusters = 1000;
	}
	
	/**
	 * Metoda vraa true ako je slika za kompresiju
	 * uspjeno uitana, inae false.
	 * 
	 * @return	da li je slika za kompresiju uitana
	 */
	public boolean loaded() {
		return loaded;
	}
	
	/**
	 * Metoda dohvaa listu alela za sliku koja sadri zadane pixele,
	 * i zadane je irine i visine.
	 * 
	 * @param pixels
	 * @param width
	 * @param height
	 * @return
	 */
	private List<PixelGene> getGeneList(int[] pixels, int width, int height, 
				Configuration config) throws InvalidConfigurationException {
		List<PixelGene> geneList = new ArrayList<PixelGene>();
		for(int i = 0; i < height; i++) {
			for(int j = 0; j < width; j++) {
				geneList.add(new PixelGene(config,
						new PixelAllele(j, i, new Color(pixels[i*width + j]))));
			}
		}
		return geneList;
	}
	
	/**
	 * Metoda generira primjerak kromosoma za trenutno uitanu sliku
	 * 
	 * @return
	 */
	private IChromosome generateSampleChromosome(Configuration config) 
									throws InvalidConfigurationException {
		if(!loaded) return null;
		if(geneList == null) throw new IllegalStateException();
		
		PixelGene[] sampleGenes = new PixelGene[geneList.size()];
		for(int i = 0; i < geneList.size(); i++) {
			sampleGenes[i] = new PixelGene(config, new PixelAllele(
					PixelAllele.COORDINATE_NOT_SET,
					PixelAllele.COORDINATE_NOT_SET, 
					Color.WHITE));
		}
		return new Chromosome(config, sampleGenes);
	}
	
	private void setup() {
		if(config == null) 
			throw new IllegalStateException("Trenutna konfiguracije je null!");
		
		System.out.println("Postavljam konfiguraciju:");
		try {
			config.setPopulationSize(populationSize);
			System.out.print("Generiram primjerak kromosoma...");
			config.setSampleChromosome(generateSampleChromosome(config));
			System.out.println("gotov");
			
			System.out.print("Stvaram ImageBreeder...");
			ImageBreeder imgBreeder = new ImageBreeder(numberOfElites);
			System.out.println("broj elitnih kromosoma postavljen na " + numberOfElites);
			
			System.out.print("Stvaram funkciju cilja...");
			ImageFitnessFunction ff = new ImageFitnessFunction();
			System.out.println("gotov");
			
			config.setFitnessFunction(ff);
			config.setBreeder(imgBreeder);
			config.getGeneticOperators().clear();
			
			System.out.print("Stvaram operator krianja 1(" + operator1Probability + "%)...");
			CrossoverOperator1 cross1 = new CrossoverOperator1(config,
					operator1Probability, false);
			System.out.println("gotov");
			System.out.print("Stvaram operator krianja 2(" + operator2Probability + "%)...");
			CrossoverOperator2 cross2 = new CrossoverOperator2(config,
					operator2Probability, false);
			System.out.println("gotov");
			System.out.print("Stvaram operator mutacije(" + mutationProbability + "%)...");
			ImageMutation mutator = new ImageMutation(config);
			System.out.println("gotov");
			
			config.addGeneticOperator(cross1);
			config.addGeneticOperator(cross2);
			config.addGeneticOperator(mutator);
			System.out.println("Konfiguracija postavljena");
		} catch(InvalidConfigurationException e) {
			
		}
	}
	
	/**
	 * Pokree kompresiju.
	 * 
	 */
	public void compress() {
		if(!loaded)
			throw new IllegalStateException("Slika nije uitana!");
		Configuration.reset();
		setup();
		
		Genotype genotype = null;
		try {
			genotype = populationConstructorC(k, config);
		} catch (InvalidConfigurationException e) {
			throw new IllegalStateException(e.getMessage());
		}
		
		solution = genotype.getFittestChromosome();
		System.out.println("Najbolja jedinka generacije " + 0 + " ima fitness: " + solution.getFitnessValue());
		
		for(int i = 1; i < numberOfGenerations; i++) {
			System.out.println("--");
			System.out.println("Poinjem novu evoluciju populacije...");
			genotype.evolve();
			solution = genotype.getFittestChromosome();
			System.out.println("Najbolja jedinka generacije " + i + " ima fitness: " + solution.getFitnessValue());
		}
		
		try {
			save(destinationImage, solution);
		} catch(IOException e) {
			System.out.println("pojavio se IO exception");
		}
	}
	
	/**
	 * Metoda vraa objekt koji je najblii zadanom alelu (tj. objektu koji predstavlja 
	 * piksel slike) iz prvih k alela u listi neiskoritenih alela. Metoda brie
	 * vraeni allel iz liste.
	 * 
	 * 
	 * @param genePool
	 * @param gene
	 * @return
	 */
	private PixelGene getNearestObject(List<PixelGene> genePool, PixelGene gene, int k) {
		if(k > genePool.size()) {
			k = genePool.size();
		}
		int smallestDistance = gene.distance(genePool.get(0)); //ImageFitnessFunction.distance(gene, genePool.get(0));
		int newDistance;
		int geneIndex = 0;
		
		for(int i = 1; i < k; i++) {
			newDistance = gene.distance(genePool.get(i));
			
			if(newDistance < smallestDistance) {
				smallestDistance = newDistance;
				geneIndex = i;
			}
		}
		
		PixelGene resultGene = genePool.get(geneIndex);
		genePool.remove(geneIndex);
		return resultGene;
	}
	
	/**
	 * Metoda stvara novi kromosom 
	 * 
	 * @param k
	 * @return
	 */
	private IChromosome chromosomeGreedyConstructor(int k, Configuration configuration) 
				throws InvalidConfigurationException {
		List<PixelGene> genePool = new ArrayList<PixelGene>(this.geneList);
		List<PixelGene> genes = new ArrayList<PixelGene>();
		
		/* Za prvi gen uzmi nasumini */
		int index = randomNumberGenerator.nextInt(genePool.size());
		PixelGene currentGene = genePool.get(index);
		genePool.remove(index);
		genes.add(currentGene);
		
		while(genePool.size() > 0) {
			currentGene = getNearestObject(genePool, currentGene, k);
			genes.add(currentGene);
		}
		
		PixelGene[] geneArray = new PixelGene[genes.size()];
		genes.toArray(geneArray);
		
		return new Chromosome(configuration, geneArray);
	}
	
	/**
	 * 
	 * @param configuration
	 * @return
	 * @throws InvalidConfigurationException
	 */
	private Genotype populationConstructorB(Configuration configuration) 
						throws InvalidConfigurationException {
		return populationConstructorC(geneList.size(), configuration);
	}
	
	/**
	 * <p>Metoda stvara novu populaciju prema zadanoj instanci objekta
	 * <code>Configuration</code> i sprema ju u novi <code>Genotype</code>
	 * objekt.</p>
	 * 
	 *  <p>Populacija se generira pohlepnim algoritmom na temelju najmanje udaljenosti
	 *  dvaju objekata (pixela). Za prvi objekt u poretku (prvi gen jedinke) uzima se 
	 *  nasumino jedan od piksela slike. Za svaki sljedei objekt odabire se onaj od 
	 *  preostalih koji ima najmanju udaljenost od prethodnog (udaljenost izmeu objekata 
	 *  dana je ImageFitness.dist metodom) iz k objekata koji jo nisu dodani u listu gena.
	 *  Ovaj nain primjenjuje se za svaku jedinku populacije.
	 *  </p>
	 * 
	 * @param k					parametar odreuje koliko se preostalih objekata razmatra
	 * 							prilikom dodavanja sljedeeg gena u jedinku
	 * @param configuration		<code>Configuration</code> objekt koji sadri postavke GA
	 * @return
	 */
	private Genotype populationConstructorC(int k, Configuration configuration)
							throws InvalidConfigurationException {
		int populationSize = configuration.getPopulationSize();
		IChromosome[] chromosomes = new IChromosome[populationSize];
		
		System.out.println("Generiram poetnu populaciju konstruktorom C");
		System.out.println("parametar k: " + k);
				
		for(int i = 0; i < populationSize; i++) {
			System.out.print("Stvaram jedinku " + i + " ...");
			chromosomes[i] = chromosomeGreedyConstructor(k, configuration);
			System.out.println("gotov");
		}
			
		System.out.println("Izgenerirao populaciju");
		Population population = new Population(configuration, chromosomes);
		return new Genotype(configuration, population);
	}
	
	/**
	 * Uitava sliku za kompresiju iz zadane datoteke.
	 * Blokira pozivajuu dretvu dok slika nije u potpunosti 
	 * uitana.
	 * 
	 * @param imageFile	<code>File</code> objekt koji pokazuje na sliku 
	 * 					koja se eli uitati 
	 */
	public void load(File imageFile) {
		if(!imageFile.exists())
			throw new IllegalArgumentException("Datoteka ne postoji!");
		
		BufferedImage image = null;
		try {
			System.out.print("Ucitavam sliku (" + imageFile.getName() + ")...");
			image = ImageIO.read(imageFile);
			System.out.println("gotov");
		} catch(IOException e) {
			throw new IllegalStateException("Nemogu ucitati sliku!");
		}
		this.width = image.getWidth();
		this.height = image.getHeight();

		int[] pixels = new int[this.width * this.height];

		pixels = image.getRGB(0, 0, this.width, this.height, pixels, 0, this.width);

		try {
			System.out.print("Dobavljam listu gena iz pixela slike...");
			this.geneList = getGeneList(pixels, width, height, config);
			System.out.println("gotov");
		} catch(InvalidConfigurationException e) {
			throw new IllegalStateException(e.getMessage());
		}

		this.loaded = true;
	}
	
	public int getWidth() {
		return width;
	}
	
	public int getHeight() {
		return height;
	}
	
	public void setCompressionPrecentage(int compressionPrecentage) {
		if(!loaded)
			throw new IllegalStateException("Slika nije uitana!");
		
		if(compressionPrecentage < 0 ) {
			throw new IllegalArgumentException("Argument mora biti pozitivanBroj");
		}
		numberOfClusters = (int)Math.ceil((double)(100 - compressionPrecentage) * this.geneList.size() / 100);
	}
	
	public int getCompressionPrecentage() {
		return (int)Math.ceil(100 * (1 - (double)numberOfClusters / this.geneList.size()));
	}
	
	private int[] toPixelArray(Gene[] genes) {
		int[] pixels = new int[genes.length];
		PixelAllele currentAllele;
		for(int i = 0; i < genes.length; i++) {
			currentAllele = (PixelAllele)genes[i].getAllele();
			pixels[i] = currentAllele.getColor().getRGB();
		}
		return pixels;
	}
	
	private List<PixelGene> getGeneList(Gene[] genes) {
		List<PixelGene> geneList = new ArrayList<PixelGene>();
		for(int i = 0; i < genes.length; i++) {
			geneList.add((PixelGene)genes[i]);
		}
		return geneList;
	}

	/**
	 * Metoda vraa greku kompresirane slike u odnosu na originalnu sliku
	 * 
	 * @param originalImagePixels
	 * @param compressedImagePixels
	 * @return
	 */
	private double calculateError(List<PixelGene> originalImagePixels, 
			Gene[] compressedImagePixels, int threshold) {
		if(originalImagePixels.size() != compressedImagePixels.length) {
			throw new IllegalArgumentException("Zadani su podaci razliitih dimenzija!");
		}
		
		long pixelErrorSum = 0;
		long maxError;
		
		for(int i = 0; i < originalImagePixels.size(); i++) {
			if(originalImagePixels.get(i).distance((PixelGene)compressedImagePixels[i]) > threshold)
				pixelErrorSum++;
		}
		maxError = originalImagePixels.size();
		return (double)pixelErrorSum / maxError;
	}
	
	/**
	 * Postavlja datoteku u koju e se spremiti kompresirana slika.
	 * 
	 * @param destinationImage
	 */
	public void setDestinationImage(File destinationImage) {
		this.destinationImage = destinationImage;
	}
	
	/**
	 * 
	 * @return
	 */
	public File getDestinationImage() {
		return destinationImage;
	}
	
	/**
	 * Sprema sliku pohranjenu u kromosom.
	 * 
	 * @param imageFile		datoteka u koju e se spremiti slika
	 * @param chromosome	kromosom koji predstavlja sliku
	 */
	private void save(File imageFile, IChromosome chromosome) throws IOException {
		if(!imageFile.exists()) {
			try {
				imageFile.createNewFile();
			} catch(IOException e) {
				throw new IOException("Nije mogue stvoriti datoteku");
			}
		}
		clusterer.setNumberOfClusters(numberOfClusters);
		Gene[] genes = clusterer.doClustering(chromosome);
		//System.out.println("genes: " + genes.length);
		
		List<PixelGene> geneList = getGeneList(genes);
		Collections.sort(geneList);
		genes = geneList.toArray(genes);
		
		System.out.print("Raunam greku kompresirane slike...");
		double error = calculateError(this.geneList, genes, errorThreshold);
		System.out.println("gotov");
		
		BufferedImage image = new BufferedImage(this.width, this.height,
				BufferedImage.TYPE_3BYTE_BGR);
		image.setRGB(0, 0, this.width, this.height, toPixelArray(genes), 0, this.width);
		
		System.out.print("Spremam sliku (" + imageFile.getName() + ")...");
		ImageIO.write(image, "PNG", imageFile);
		System.out.println("gotov");
		
		System.out.println("--");
		System.out.println("Stupanj kompresije slike: " + this.getCompressionPrecentage() + "%");
		System.out.println("Greka kompresirane slike: " + error * 100 + "%");
	}
}
