package hr.fer.ga.imgcompression;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import org.jgap.Chromosome;
import org.jgap.Configuration;
import org.jgap.Gene;
import org.jgap.Genotype;
import org.jgap.IChromosome;
import org.jgap.IGeneticOperatorConstraint;
import org.jgap.IUniversalRateCalculator;
import org.jgap.InvalidConfigurationException;
import org.jgap.Population;
import org.jgap.RandomGenerator;
import org.jgap.impl.MutationOperator;

/**
 * 
 */

/**
 * @author Ivan Kut
 *
 */
public class ImageMutation extends MutationOperator {

	/**
	 * vjerojatnost krianja zadana brojem jedinki koje
	 * e se kriati (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private int m_crossoverRate;
	
	/**
	 * vjerojatnost krianja zadana postotkom jedinki populacije
	 * (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private double m_crossoverRatePercent;

	/**
	 * kalkulator za dinamiko odreivanje vjerojatnosti krianja
	 * (iskljuivo je s m_crossoverRatePrecent i m_crossoverRateCalc)
	 */
	private IUniversalRateCalculator m_crossoverRateCalc;
	
	/**
	 * ako je vrijednost postavljena na true kria i novo stvorene kromosome iz
	 * populacije potomaka (kromosome stare 0 godina 
	 */
	private boolean m_xoverNewAge;
	
	public ImageMutation()
		throws InvalidConfigurationException {
		super(Genotype.getStaticConfiguration());
		init();
	}
	
	public ImageMutation(final Configuration config) throws InvalidConfigurationException {
		super(config);
		init();
	}
		
	protected void init() {
	    m_crossoverRate = 6;
	    m_crossoverRatePercent = -1;
	    setMutationRateCalc(null);
	    setMutateNewAge(true);
	}
	
	public ImageMutation(final Configuration a_configuration,
				final IUniversalRateCalculator a_crossoverRateCalculator)
										throws InvalidConfigurationException {
		super(a_configuration);
		setMutationRateCalc(a_crossoverRateCalculator);
	}
	
	/**
	 * Postavlja kalkulator za dinamiko odreivanje vjerojatnosti krianja
	 * 
	 * @param a_crossoverRateCalculator
	 */
	public void setMutationRateCalc(final IUniversalRateCalculator
											a_crossoverRateCalculator) {
		m_crossoverRateCalc = a_crossoverRateCalculator;
		if (a_crossoverRateCalculator != null) {
			m_crossoverRate = -1;
			m_crossoverRatePercent = -1d;
		}
	}
	
	public ImageMutation(final Configuration a_configuration,
			final double a_crossoverRatePercentage,
			final boolean a_xoverNewAge)
				throws InvalidConfigurationException {
		super(a_configuration);
		
		if (a_crossoverRatePercentage <= 0.0d) {
			throw new IllegalArgumentException("Crossover rate must be greater zero");
		}
		m_crossoverRatePercent = a_crossoverRatePercentage;
		m_crossoverRate = -1;
		setMutationRateCalc(null);
		setMutateNewAge(a_xoverNewAge);
	}
	

	public void setMutateNewAge(boolean a_xoverNewAge) {
		m_xoverNewAge = a_xoverNewAge;
	}

	
	/**
	 * Vraa gene kromosoma pospremljene u ArrayListu
	 * 
	 * @param chrom
	 * @return
	 */
	private List<PixelGene> getGeneList(IChromosome chrom) {
		List<PixelGene> geneList = new ArrayList<PixelGene>();
		Gene[] genes = chrom.getGenes();
		for(int i = 0; i < genes.length; i++)
			geneList.add((PixelGene)genes[i]);
			
		return geneList;
	}
	
	public IChromosome mutate(IChromosome chromosome) 
					throws InvalidConfigurationException {
		System.out.print("Mutiram kromosom  " + chromosome.getFitnessValue() + " ...");
		Configuration config = getConfiguration();
		int index1 = getConfiguration().getRandomGenerator().nextInt(chromosome.size());
		int index2 = getConfiguration().getRandomGenerator().nextInt(chromosome.size());
		
		List<PixelGene> geneList = getGeneList(chromosome);
		PixelGene temporaryGeneStorage = geneList.get(index1);
		geneList.set(index1, geneList.get(index2));
		geneList.set(index2, temporaryGeneStorage);
		PixelGene[] genes = new PixelGene[chromosome.size()];
		genes = geneList.toArray(genes);
		System.out.println("gotov");
		return new Chromosome(config, genes);
	}
	
	  /**
	   * @return a_xoverNewAge true: also x-over chromosomes with age of zero (newly
	   * created chromosomes)
	   *
	   * @author Klaus Meffert
	   * @since 3.3.2
	   */
	  public boolean isMutateNewAge() {
	    return m_xoverNewAge;
	  }
	
	/**
	 * Metoda vraa broj krianja koji je potrebno obaviti
	 * prema trenutnim postavkama.
	 * 
	 * @param populationSize
	 * @return
	 */
	private int getNumberOfCrossovers(int size) {
	    int numCrossovers = 0;
	    if (m_crossoverRate >= 0) {
	    	numCrossovers = size / m_crossoverRate;
	    } else if (m_crossoverRateCalc != null) {
	    	numCrossovers = size / m_crossoverRateCalc.calculateCurrentRate();
	    } else {
	    	numCrossovers = (int) (size * m_crossoverRatePercent);
	    }
	    return numCrossovers;
	}
	
	/**
	 * Ova metoda zove se jednom po evoluciji: mora evoluirat sve
	 * 
	 * kod evolucije: prvo zove Operatore krianja (genetic operators)
	 * onda zove prirodne selektore
	 * 
	 * @param	population 	stara populacija
	 * @param	chromosomes		izgenerirani potomci se stavljaju tu
	 */
	@Override
	public void operate(Population population, List chromosomes) {
		int size = Math.min(getConfiguration().getPopulationSize(),
                population.size());
		int numCrossovers = getNumberOfCrossovers(population.size());
	    RandomGenerator generator = getConfiguration().getRandomGenerator();
	    IGeneticOperatorConstraint constraint = getConfiguration().
	        getJGAPFactory().getGeneticOperatorConstraint();
	    	    
	    int index;
	    for (int i = 0; i < numCrossovers; i++) {
	    	index = generator.nextInt(size);
	    	IChromosome parent = population.getChromosome(index);

	    	if (!isMutateNewAge() && parent.getAge() < 1) {
	    		continue;
	    	}
	    	if (constraint != null) {
	    		List v = new Vector();
	    		v.add(parent);
	        	if (!constraint.isValid(population, v, this)) {
	        		/* zabranjeno je ovo mutiranje */
	        		continue;
	        	}
	    	}
	    	
	    	IChromosome child = null;
	    	try {
	    		child = mutate(parent);
	    	} catch(InvalidConfigurationException e) {
	    		throw new IllegalStateException();
	    	}
	    	chromosomes.add(child);
	    }
	}
}
